import os
import json
import tempfile
from datetime import timedelta, datetime
from pathlib import Path
import time
from typing import Dict, List
import unicodedata
import re

from croniter import croniter
from ruxit.api.base_plugin import RemoteBasePlugin
from ruxit.api.exceptions import NothingToReportException, ConfigException

from dbquery_imports.log_proxy import LogProxy
from dbquery_imports.models import Result, QueryConfig
from dbquery_imports import connector
from dbquery_imports.environment import get_api_url


class CustomDBQueryPluginRemote(RemoteBasePlugin):
    def initialize(self, **kwargs):
        self.validate_parameters()
        self.temp_folder = create_or_get_temp_folder()


    def query(self, **kwargs):
        start_time = time.time()

        self.logger.setLevel(self.config.get("log_level"))

        # These are to tie the metrics to a custom device
        # Useful for management zones and alerting
        group_name = self.config.get("group_name") or "Custom DB Queries Group"
        device_name = self.config.get("custom_device_name") or "Custom DB Queries Device"
        group = self.topology_builder.create_group(group_name, group_name)
        self.custom_device = group.create_device(device_name)
        custom_device_id = f"CUSTOM_DEVICE-{self.custom_device.id:X}"

        # This is for us to send statistics about the queries, response time, status, number of rows
        self.send_statistics = self.config.get("send_statistics", False)
        statistics_group = self.topology_builder.create_group("DBQuery Monitoring", "DBQuery Monitoring")
        name = self.activation.endpoint_name or self.activation.entity_id
        self.statistics_device = statistics_group.create_device(
            f"Statistics ({self.activation.entity_id})", f"DB Query Statistics ({name})"
        )

        if self.send_statistics:
            self.statistics_device.report_property("Endpoint name or ID", f"{name}")
            self.statistics_device.report_property("Type", self.config.get("database_type"))
            self.statistics_device.report_property("Host", self.config.get("database_host"))

        # First, figure out which queries to run, based on the schedules
        queries = self.get_queries()
        queries_to_run = [q for q in queries if q.will_run_now]

        if self.send_statistics:
            self.statistics_device.report_property("Queries", f"{len(queries)}")
        if queries_to_run:
            # Write the config only for the queries that must run now, and start the java process that does the querying
            self.write_config(queries_to_run)

            # Extra arguments to pass to the JVM, like -Xmx60m
            jvm_extra_args = []
            if self.config.get("jvm_extra_args"):
                jvm_extra_args = self.config.get("jvm_extra_args").splitlines()

            # Spawn the java process that will run the queries
            # The password is passed as an Environment Variable for security
            env = os.environ.copy()
            env.update({
                "DATABASE_PASSWORD": self.config["database_password"],
                "DYNATRACE_CUSTOM_DEVICE_ID": custom_device_id,
                "DYNATRACE_API_URL": get_api_url(),
                "DYNATRACE_API_TOKEN": self.config["dt_api_token"],
                "DYNATRACE_TEMP_FOLDER": f"{self.temp_folder}",
                "DYNATRACE_ENDPOINT": f"{self.activation.entity_id}",
            })
            connector.query(
                self.activation.entity_id,
                self.temp_folder,
                env=env,
                java_path=self.config.get("jvm_path"),
                jvm_extra_args=jvm_extra_args,
            )

        self.parse_logs()
        self.custom_device.absolute("duration", time.time() - start_time)

    def parse_logs(self):
        log = Path(self.temp_folder, "log", f"{self.activation.entity_id}.log")
        cache = Path(self.temp_folder, f"{self.activation.entity_id}.json")

        for line in LogProxy(log, cache).get_lines_from_file():
            self.logger.info(line)
            if " ERROR " in line or " WARN " in line or " FATAL " in line:
                self.statistics_device.report_custom_info_event(
                    f"Received a bad line from the log: {line}", "Bad log line received", properties={"line": line}
                )

    def get_queries(self) -> List[QueryConfig]:
        """
        Determines which queries need to be run on this interval
        This is based on the schedule of the query, by default a query runs every minute
        :return: A list of parsed QueryConfig objects
        """
        queries = []
        now = datetime.now().astimezone()

        # We have up to 10 queries
        for i in range(1, 11):
            name = self.config.get(f"query_{i}_name", "")

            try:
                if not name:
                    # Queries that don't have a name are ignored
                    continue

                # Default schedule is every minute
                schedule = self.config.get(f"query_{i}_schedule") or "*/1 * * * *"

                query_string = self.config[f"query_{i}_value"]

                # The user can specify key=value dimension pairs
                extra_dimensions = self.config.get(f"query_{i}_extra_dimensions", "") or []
                if extra_dimensions:
                    extra_dimensions = [d.strip() for d in extra_dimensions.split(",")]

                # The user can specify column names to capture as values
                value_columns = []
                value_columns_raw = self.config.get(f"query_{i}_value_columns", "")
                if value_columns_raw:
                    value_columns = value_columns_raw.split(",")

                # The user can specify column names to capture as dimensions
                dimension_columns_raw = self.config.get(f"query_{i}_dimension_columns", "")
                dimension_columns = []
                if dimension_columns_raw:
                    dimension_columns = dimension_columns_raw.split(",")

                # Parses the cron schedule for this query
                cron = croniter(schedule, now, ret_type=datetime)
                previous_execution = cron.get_prev().astimezone()
                next_execution = cron.get_next().astimezone()

                query_config = QueryConfig(
                    name, query_string, value_columns, dimension_columns, extra_dimensions=extra_dimensions, schedule=schedule
                )

                # Figure out if this query needs to run now
                time_since = now - previous_execution
                if time_since <= timedelta(minutes=1):
                    self.logger.info(
                        f"Adding query '{name}' because it has been scheduled to run {time_since} ago ({previous_execution}) "
                    )
                    query_config.will_run_now = True
                else:
                    self.logger.info(f"Not adding query '{name}'. Next execution at {next_execution}, now is {now}")
                    query_config.will_run_now = False

                queries.append(query_config)
                if self.send_statistics:
                    state = "RUNNING" if query_config.will_run_now else "CACHED"
                    self.statistics_device.state_metric("query_state", state, dimensions={"Query": query_config.name})

            except Exception as e:
                self.logger.error(f"Could not add query {name}, error: {e}")

        return queries

    def build_connection_string(self) -> str:
        if self.config.get("jdbc_connection_string", ""):
            return self.config.get("jdbc_connection_string")

        database_type = self.config["database_type"]
        host = self.config["database_host"]
        port = self.config["database_port"]
        database = self.config["database_name"]

        if database_type == "MySQL":
            return f"jdbc:mysql://{host}:{port}/{database}"

        if database_type == "PostgreSQL":
            return f"jdbc:postgresql://{host}:{port}/{database}"

        if database_type == "Oracle":
            if self.config["oracle_listener_type"] == "SID":
                return f"jdbc:oracle:thin:@{host}:{port}:{database}"
            else:
                return f"jdbc:oracle:thin:@//{host}:{port}/{database}"

        if database_type == "SQL Server":
            return f"jdbc:sqlserver://{host}:{port};databaseName={database}"

        if database_type == "DB2":
            return f"jdbc:db2://{host}:{port}/{database}"

        if database_type == "Informix":
            return f"jdbc:informix-sqli://{host}:{port}/{database}"

        if database_type == "SAP HANA":
            return f"jdbc:sap://{host}:{port}/?encrypt=true&validateCertificate=false"

    def write_config(self, queries_to_run: List[QueryConfig]):
        """
        Writes the database configuration to a json file named after the endpoint ID
        """

        driver_classes = {
            "DB2": "com.ibm.db2.jcc.DB2Driver",
            "SQL Server": "com.microsoft.sqlserver.jdbc.SQLServerDriver",
            "Oracle": "oracle.jdbc.driver.OracleDriver",
            "MySQL": "com.mysql.cj.jdbc.Driver",
            "PostgreSQL": "org.postgresql.Driver",
            "Informix": "com.informix.jdbc.IfxDriver",
            "SAP HANA": "com.sap.db.jdbc.Driver",
        }

        queries_json = [query.__dict__ for query in queries_to_run]
        config = {
            "username": self.config["database_username"],
            "logLevel": self.config["log_level"],
            "queries": queries_json,
            "driverClassName": driver_classes.get(self.config["database_type"]),
            "connectionString": self.build_connection_string(),
        }
        self.logger.info(f"Writing config {config}")

        path = Path(self.temp_folder, "config")
        with open(Path(path, f"{self.activation.entity_id}.json"), "w") as f:
            json.dump(config, f, indent=2)

    def parse_results(self):
        result = self.read_results_from_file()
        self.logger.info(f"Processing result {result}")

        if result.error:
            # This means there was an error connecting to the Database
            # Report the error and the availability
            error_message = f"Error running queries: {result.error_message}"
            self.logger.error(error_message)
            self.custom_device.report_error_event(error_message, "Custom DB Query Extension error")
            if self.send_statistics:
                self.statistics_device.state_metric("result_state", "ERROR")

        else:
            if self.send_statistics:
                self.statistics_device.state_metric("result_state", "RUNNING")
            # If we get here we have successfully queried the database, parse the results

    def read_results_from_file(self) -> Result:
        """
        Reads the <endpoint_id>.json file inside the results folder
        and parses the data into a models.Result object
        """
        results_file = Path(self.temp_folder, "results", f"{self.activation.entity_id}.json")
        if results_file.exists():
            with open(results_file, "r", errors="replace") as f:
                try:
                    return Result.from_json(json.load(f))
                except Exception as e:
                    self.logger.error(f"Could not parse json from {results_file}: {e}")
                    raise NothingToReportException(f"Error: {e}")
        return Result()

    def validate_parameters(self):

        # Check that at least one query was configured
        query_names = [self.config.get(f"query_{i}_name", "") for i in range(1, 11)]
        # Count the number of query names that are not empty
        num_queries = len([q for q in query_names if q])
        self.logger.info(f"{num_queries} queries configured")
        if num_queries == 0:
            raise ConfigException("At least one query must be configured")

        token = self.config.get("dt_api_token", "")
        if not token:
            raise ConfigException("API Token must be set")


def chunks(elements, size):
    """
    Breaks a list into smaller lists of size elements
    """
    for i in range(0, len(elements), size):
        yield elements[i : i + size]


def slugify(value: str) -> str:
    """
    Converts a string into an url safe string
    """
    value = unicodedata.normalize("NFKC", value)
    value = re.sub(r"[:.-/\s]+", "_", value)
    value = re.sub(r"[^\w\s-]", "", value)
    return value.lower()


def create_or_get_temp_folder():
    temp_folder = Path(tempfile.gettempdir(), "dynatrace-ag-dbquery")
    if not temp_folder.exists():
        os.mkdir(temp_folder)

    other_paths = [
        Path(temp_folder, "config"),
        Path(temp_folder, "log"),
        Path(temp_folder, "results"),
    ]

    for path in other_paths:
        if not path.exists():
            os.mkdir(path)

    return temp_folder
